/*******************************************************************************
!   Copyright(C) 2011-2012 Intel Corporation. All Rights Reserved.
!
!   The source code, information  and  material ("Material") contained herein is
!   owned  by Intel Corporation or its suppliers or licensors, and title to such
!   Material remains  with Intel Corporation  or its suppliers or licensors. The
!   Material  contains proprietary information  of  Intel or  its  suppliers and
!   licensors. The  Material is protected by worldwide copyright laws and treaty
!   provisions. No  part  of  the  Material  may  be  used,  copied, reproduced,
!   modified, published, uploaded, posted, transmitted, distributed or disclosed
!   in any way  without Intel's  prior  express written  permission. No  license
!   under  any patent, copyright  or  other intellectual property rights  in the
!   Material  is  granted  to  or  conferred  upon  you,  either  expressly,  by
!   implication, inducement,  estoppel or  otherwise.  Any  license  under  such
!   intellectual  property  rights must  be express  and  approved  by  Intel in
!   writing.
!
!   *Third Party trademarks are the property of their respective owners.
!
!   Unless otherwise  agreed  by Intel  in writing, you may not remove  or alter
!   this  notice or  any other notice embedded  in Materials by Intel or Intel's
!   suppliers or licensors in any way.
!
!*******************************************************************************
!   Content:
!       SGEMM, DGEMM, CGEMM, ZGEMM benchmarks driver
!******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "bench.h"
#include "utils.h"
#include "cross_timer.h"

#ifdef __INTEL_OFFLOAD
#include "offload.h"
#endif

#ifndef _WINHOST
TARGET_MIC_PUSH
#else
#pragma offload_attribute(push,target(mic))
#endif
#include "omp.h"
#ifndef _WINHOST
TARGET_MIC_POP
#else
#pragma offload_attribute(pop)
#endif

int coprocessor = 0;
static size_t elemsA, elemsB, elemsC;
TARGET_MIC_ATTR static fptype_t *A, *B, *C;
cross_timer_t timeIn, timeOut;

void initialize_bench(int maxM, int maxN, int maxK, int num_threads,
		int coproc)
{
	coprocessor = coproc;
	elemsA = (MIN(maxM, maxK) + FIX_LD_SHFT);
	elemsA *= MAX(maxM, maxK);
	elemsB = (MIN(maxN, maxK) + FIX_LD_SHFT);
	elemsB *= MAX(maxN, maxK);
	elemsC = (maxM + FIX_LD_SHFT);
	elemsC *= maxN;

	A = bench_malloc(elemsA * sizeof(fptype_t));
	B = bench_malloc(elemsB * sizeof(fptype_t));
	C = bench_malloc(elemsC * sizeof(fptype_t));

	fill_matrix(A, elemsA);
	fill_matrix(B, elemsB);
	fill_matrix(C, elemsC);

#ifdef __INTEL_OFFLOAD
#pragma offload target(mic:coprocessor) \
	in(num_threads) \
	in(A:length(elemsA) alloc_if(1) free_if(0)) \
	in(B:length(elemsB) alloc_if(1) free_if(0)) \
	in(C:length(elemsC) alloc_if(1) free_if(0))
#endif
	{
		int max_threads = kmp_get_affinity_max_proc();
		if (getenv("OMP_NUM_THREADS") == NULL || num_threads > 0) {
			if (num_threads <= 0) {
#if defined(__MIC__) && defined(__INTEL_OFFLOAD)
				/* Do not use the last 4 threads on the coprocessor in the
				 * offload mode since they are busy. */
				num_threads = max_threads - 4;
#else
				num_threads = max_threads;
#endif
				printf("threads used: %d (autodetected)\n", num_threads);
			} else {
				printf("threads used: %d (set by user)\n", num_threads);
			}
			fflush(NULL);
			omp_set_num_threads(num_threads);
		} else {
			printf("threads used: %d (OMP_NUM_THREADS)\n",
					omp_get_max_threads());
			fflush(NULL);
		}

#ifdef __MIC__
		if (getenv("KMP_AFFINITY") == NULL) {
#pragma omp parallel num_threads(num_threads)
			{
				int ithr = omp_get_thread_num();
				if (ithr < max_threads - 4)
					ithr += 1;
				else if (ithr == max_threads - 4)
					ithr = 0;
				kmp_affinity_mask_t m;
				kmp_create_affinity_mask(&m);
				kmp_set_affinity_mask_proc(ithr, &m);
				kmp_set_affinity(&m);
				kmp_destroy_affinity_mask(&m);
			}
			printf("affinity    : flat, set by benchmark\n");
			fflush(NULL);
		} else {
			printf("affinity    : KMP_AFFINITY\n");
			fflush(NULL);
		}
#else
			printf("affinity    : KMP_AFFINITY (if any)\n");
			fflush(NULL);
#endif
	}
}

void finalize_bench()
{
#ifdef __INTEL_OFFLOAD
#pragma offload target(mic:coprocessor) \
	nocopy(A, B, C : alloc_if(0) free_if(1))
	{ }
#endif
	bench_free(A, elemsA * sizeof(fptype_t));
	bench_free(B, elemsB * sizeof(fptype_t));
	bench_free(C, elemsC * sizeof(fptype_t));
}

double xgemm_bench(char transa, char transb,
		int M, int N, int K, fptype_t alpha, fptype_t beta)
{
	int AisN = (transa == 'N' || transa == 'n');
	int BisN = (transb == 'N' || transb == 'n');
	int colsA = (AisN ? K : M);
	int colsB = (BisN ? N : K);
	int colsC = N;
	int LDA = FIX_LD(AisN ? M : K);
	int LDB = FIX_LD(BisN ? K : N);
	int LDC = FIX_LD(M);

	cross_timer_sample(&timeIn);

#if defined(__INTEL_OFFLOAD)
#pragma offload target(mic:coprocessor) \
	in(A:length(LDA * colsA) alloc_if(0) free_if(0)) \
	in(B:length(LDB * colsB) alloc_if(0) free_if(0)) \
	in(C:length(LDC * colsC) alloc_if(0) free_if(0))
	{ }
#endif

#ifdef __INTEL_OFFLOAD
#pragma offload target(mic:coprocessor) \
	nocopy(A, B, C : alloc_if(0) free_if(0)) \
	in(transa, transb, M, N, K, alpha, LDA, LDB, beta, LDC)
#endif
	{
		xgemm(&transa, &transb, &M, &N, &K, &alpha,
				A, &LDA, B, &LDB, &beta, C, &LDC);
	}

#if defined(__INTEL_OFFLOAD)
#pragma offload target(mic:coprocessor) \
	nocopy(A, B : alloc_if(0) free_if(0)) \
	out(C : length(LDA * colsC) alloc_if(0) free_if(0))
	{ }
#endif
	cross_timer_sample(&timeOut);
	return cross_timer_diff(timeIn, timeOut);
}

