/*******************************************************************************
!   Copyright(C) 2011-2012 Intel Corporation. All Rights Reserved.
!
!   The source code, information  and  material ("Material") contained herein is
!   owned  by Intel Corporation or its suppliers or licensors, and title to such
!   Material remains  with Intel Corporation  or its suppliers or licensors. The
!   Material  contains proprietary information  of  Intel or  its  suppliers and
!   licensors. The  Material is protected by worldwide copyright laws and treaty
!   provisions. No  part  of  the  Material  may  be  used,  copied, reproduced,
!   modified, published, uploaded, posted, transmitted, distributed or disclosed
!   in any way  without Intel's  prior  express written  permission. No  license
!   under  any patent, copyright  or  other intellectual property rights  in the
!   Material  is  granted  to  or  conferred  upon  you,  either  expressly,  by
!   implication, inducement,  estoppel or  otherwise.  Any  license  under  such
!   intellectual  property  rights must  be express  and  approved  by  Intel in
!   writing.
!
!   *Third Party trademarks are the property of their respective owners.
!
!   Unless otherwise  agreed  by Intel  in writing, you may not remove  or alter
!   this  notice or  any other notice embedded  in Materials by Intel or Intel's
!   suppliers or licensors in any way.
!
!*******************************************************************************
!   Content:
!       SGEMM, DGEMM, CGEMM, ZGEMM benchmarks
!******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#ifndef _WIN32
#include <getopt.h>
#else
#include "getopt_win.h"
#endif

#include "utils.h"
#include "bench.h"

void usage(char *progname)
{
	printf("\n"
		"Usage: %s [options...]\n\n"
		"Supported options are:\n"
		"  -n <# of threads> : default value = -1 (autodetect)\n"
		"  -T <target coprocessor> : default value = -1 (autodetect)\n"
		"  -i <minimum # of bench iterations> : default value = %d\n"
		"  -t <minimum total bench time> : default value = %f sec.\n"
		"  -f <initial size of testing matrix> : default value = %d\n"
		"  -l <final size of testing matrix> : default value = %d\n"
		"  -s <size step> : default value = %d\n"
		"  -m <transposition mode> : 0 = NN, 1 = NT, 2 = TN, 3 = TT;\n"
		"                            default: run all modes\n\n",
		progname,
		DEF_MIN_NITERS, DEF_MIN_T, DEF_FIRST_IND, DEF_LAST_IND, DEF_STEP);
	exit(1);
}

#define STRINGIFY1(x) #x
#define STRINGIFY(x) STRINGIFY1(x)

int main(int argc, char *argv[])
{
	MKLVersion Version;
	int i;
	int mode;
	int firstInd = DEF_FIRST_IND;
	int lastInd = DEF_LAST_IND;
	int step = DEF_STEP;
	int min_niters = DEF_MIN_NITERS;
	double min_t = DEF_MIN_T;
	int num_threads = DEF_NUM_THREADS;
	int coprocessor = DEF_COPROCESSOR;
	int min_mode = 0, max_mode = 3;

	int c;
	while ((c = getopt(argc, argv, "n:f:l:s:i:d:T:t:m:h")) != -1)
		switch (c) {
		case 'f': firstInd = atoi(optarg); break;
		case 'l': lastInd = atoi(optarg); break;
		case 's': step = atoi(optarg); break;
		case 'i': min_niters = atoi(optarg); break;
		case 't': min_t = atof(optarg); break;
#ifdef __INTEL_OFFLOAD
		case 'T': coprocessor = atoi(optarg); break;
		case 'd': coprocessor = atoi(optarg); fprintf(stderr, "WARNING:  -d option has been deprecated, use -T instead\n"); break;
#endif
		case 'n': num_threads = atoi(optarg); break;
		case 'm': min_mode = max_mode = atoi(optarg); break;
		case 'h': usage(argv[0]); break;
		}

	/* Check loop bounds for mode */
	min_mode = (min_mode < 0) ? 0 : min_mode;
	max_mode = (max_mode < 0) ? 0 : max_mode;
	min_mode = (min_mode > 3) ? 3 : min_mode;
	max_mode = (max_mode > 3) ? 3 : max_mode;
	max_mode = (max_mode < min_mode) ? min_mode : max_mode;
	/* Check loop bounds for matrix size */
	firstInd = (firstInd < 1) ? 1 : firstInd;
	lastInd = (lastInd < 1) ? 1 : lastInd;
	firstInd = (firstInd > MAX_ALLOWED_IND) ? MAX_ALLOWED_IND : firstInd;
	lastInd = (lastInd > MAX_ALLOWED_IND) ? MAX_ALLOWED_IND : lastInd;
	lastInd = (lastInd < firstInd) ? firstInd : lastInd;
	step = (step < 1) ? 1 : step;
	step = (step > MAX_ALLOWED_IND) ? MAX_ALLOWED_IND : step;

	printf("benchmarking: " STRINGIFY(xgemm) "\n");
#ifdef __INTEL_OFFLOAD
	printf("timer       : full\n");
	printf("coprocessor : %d\n", coprocessor);
#else
	printf("timer       : native\n");
#endif
	printf("num_threads : %d%s\n", num_threads,
			(num_threads < 0 ? " (autodetect / environment)" : ""));
	printf("min_niters  : %d\n", min_niters);
	printf("min_t       : %f\n", min_t);
	printf("first index : %d\n", firstInd);
	printf("last  index : %d\n", lastInd);
	printf("step        : %d\n", step);
	printf("data transf.: %s\n",
#if defined(__INTEL_OFFLOAD)
			"on"
#elif defined(__MIC__)
			"off"
#else
			"maybe (depends on MKL AO setting)"
#endif
			);


	mkl_get_version(&Version);
	printf("MKL         : %d.%d.%d build %s (%s)\n",
			Version.MajorVersion, Version.MinorVersion, Version.UpdateVersion,
			Version.Build, Version.ProductStatus);
	printf("processor   : %s\n", Version.Processor);
	printf("CPU freq.   : %.2f (may float due to scaling)\n",
			mkl_get_cpu_frequency());
	printf("max threads : %d\n", mkl_get_max_threads());
#ifndef __MIC__
	printf("# of co-proc: %d\n",
#ifdef __INTEL_OFFLOAD
			_Offload_number_of_devices()
#else
			mkl_mic_get_device_count()
#endif
		);
#endif

	initialize_bench(lastInd, lastInd, lastInd, num_threads, coprocessor);

	for (mode = min_mode; mode < max_mode + 1; mode++) {
		int best_i = firstInd;
		double best_avg_gflops = 0;
		double best_min_gflops = 0;
		double best_max_gflops = 0;
		double best_stddev = 0;
		char transa, transb;

		transa = (mode / 2 == 0 ? 'N' : 'T');
		transb = (mode % 2 == 0 ? 'N' : 'T');
		printf("\n#%d: %c%c\n", mode, transa, transb);

		printf("\ntesting XGEMM( '%c', '%c', n, n, ... )\n\n",
				transa, transb);
		printf("%11s %10s %10s %10s %10s\n",
				"n", "min", "avg", "max", "stddev");
		fflush(NULL);

		for (i = firstInd; i <= lastInd; i += step) {
#if defined(SINGLE_PREC) || defined (DOUBLE_PREC)
			fptype_t alpha = 1.0;
			fptype_t beta = 1.0;
#else
			fptype_t alpha = {1.0, 0.0};
			fptype_t beta = {1.0, 0.0};
#endif
			int iter = 0;
			double min = DBL_MAX, max = -DBL_MAX, sum = 0.0, stddev = 0.0;
			int M = i, N = i, K = i;
			double avg;
			double ngops;
			double min_gflops;
			double avg_gflops; 
			double max_gflops; 
			volatile double t;

			/* Two warmup calls (aids stability on small sizes) */
			t = xgemm_bench(transa, transb, M, N, K, alpha, beta);
			t = xgemm_bench(transa, transb, M, N, K, alpha, beta);

			/* Run computations until either the timeout or iterations limit
			 * is reached */

			while ((iter < min_niters || sum < min_t)) {
				double t = xgemm_bench(transa, transb, M, N, K, alpha, beta);
				min = (min < t) ? min : t;
				max = (max > t) ? max : t;
				sum += t;
				stddev += t * t;
				iter++;
			}
			avg = sum / iter;
			stddev = sqrt(stddev / iter - avg * avg);

			ngops = 2.0 * M * N * K / 1E9;
#if !defined(SINGLE_PREC) && !defined (DOUBLE_PREC)
			ngops *= 4.0;
#endif
			min_gflops = ngops / max;
			avg_gflops = ngops / avg;
			max_gflops = ngops / min;
			printf("%11d %10.2f %10.2f %10.2f %10.3e\n",
					i, min_gflops, avg_gflops, max_gflops, stddev);
			fflush(NULL);

			if (best_avg_gflops < avg_gflops) {
				best_i = i;
				best_avg_gflops = avg_gflops;
				best_min_gflops = min_gflops;
				best_max_gflops = max_gflops;
				best_stddev = stddev;
			}
		}

		printf("*%10d %10.2f %10.2f %10.2f %10.3e\n", best_i,
				best_min_gflops, best_avg_gflops,
				best_max_gflops, best_stddev);
		fflush(NULL);
	}

	finalize_bench();

	return 0;
}

