# Copyright 2012, 2013, 2014 Intel Corporation All Rights Reserved.
#
# This software is supplied under the terms of a license
# agreement or nondisclosure agreement with Intel Corp.
# and may not be copied or disclosed except in accordance
# with the terms of that agreement.\
#
#  Author:  Christopher M. Cantalupo

"""
Module containing the Info class which follows the mono-state
pattern. This class stores information about the hardware and software
characteristics of a system.  There are also module level functions
for accessing the system information that mirror the Info class API.
"""

import subprocess
import re
import hashlib
import shlex
import os
import distutils.version
import copy
import platform
import sys

import common as micp_common
import connect as micp_connect


def system_hw_hash():
    return Info().system_hw_hash()

def mpss_version():
    return Info().mpss_version()

def micinfo():
    return Info().micinfo()

def num_cores(devIdx):
    info = Info()
    origIdx = info.get_device_index()
    info.set_device_index(devIdx)
    result = info.num_cores()
    info.set_device_index(origIdx)
    return result

def gddr_size(devIdx):
    info = Info()
    origIdx = info.get_device_index()
    info.set_device_index(devIdx)
    result = info.gddr_size()
    info.set_device_index(origIdx)
    return result

def mic_stepping(devIdx):
    info = Info()
    origIdx = info.get_device_index()
    info.set_device_index(devIdx)
    result = info.mic_stepping()
    info.set_device_index(origIdx)
    return result

def micinfo_basic(devIdx):
    info = Info()
    origIdx = info.get_device_index()
    info.set_device_index(devIdx)
    result = info.micinfo_basic()
    info.set_device_index(origIdx)
    return result

def mic_sku(devIdx):
    info = Info()
    origIdx = info.get_device_index()
    info.set_device_index(devIdx)
    result = info.mic_sku()
    info.set_device_index(origIdx)
    return result

class Borg:
    _sharedState = {}
    def __init__(self):
        self.__dict__ = self._sharedState

class Info(object, Borg):
    """
    Class which holds the output from calls to deterimine the system
    configuration.  Also holds a MicInfoDict in attribute micinfo.
    """
    def __init__(self):
        Borg.__init__(self)
        if self.__dict__ == {}:
            self._devIdx = 0
            self._devIP = {}
            self._init_command_list()
            self._init_command_dict()
            self._init_micinfo_dict()

    def __str__(self, categories = None):
        if not categories:
            categories = self.get_app_list()
        elif type(categories) == str:
            categories = [categories]
        unorderedKeys = self._commandDict.keys()
        keys = []
        for category in categories:
            theseKeys = sorted([key for key in unorderedKeys if category in key])
            if len(theseKeys) == 0:
                raise LookupError('Requested application {0} not found'.format(category))
            keys.extend(theseKeys)
        result = []
        keys = sorted(list(set(keys)))
        for cmd in keys:
            result.append(micp_common.star_border(cmd))
            result.append(self._commandDict[cmd])
        return '\n'.join(result)

    def get_device_index(self):
        return self._devIdx

    def set_device_index(self, devIdx):
        self._devIdx = devIdx

    def get_device_name(self):
        if self._devIdx == -1:
            return 'localhost'
        else:
            return 'mic{0}'.format(self.get_device_index())

    def set_device_name(self, name):
        if name == 'localhost':
            self._devIdx = -1
        elif name.startswith('mic'):
            self._devIdx = int(name[3:])
        else:
            raise NameError('Unknown device name {0}'.format(name))

    def num_cores(self):
        devID = self.get_device_name()
        return int(self._micinfoDict[devID]['Total No of Active Cores'])

    def gddr_size(self):
        devID = self.get_device_name()
        size, units = tuple(self._micinfoDict[devID]['GDDR Size'].split())
        if units.lower() == 'b':
            factor = 1
        elif units.lower() == 'kb':
            factor = 1024
        elif units.lower() == 'mb':
            factor = 1024**2
        elif units.lower() == 'gb':
            factor = 1024**3
        else:
            raise NameError('Unknown units for micinfo GDDR Size {0}'.format(units))
        return int(float(size)*factor)

    def mic_stepping(self):
        devID = self.get_device_name()
        try:
            result = self._micinfoDict[devID]['Coprocessor Stepping']
        except KeyError:
            result = 'NotAvailble'
        return result

    def mic_sku(self):
        devID = self.get_device_name()
        try:
            result = self._micinfoDict[devID]['Board SKU']
        except KeyError:
            result = 'NotAvailable'
        return result

    def micinfo(self):
        return self.__str__('micinfo')

    def micinfo_basic(self):
        devID = self.get_device_name()
        try:
            result = 'Core Freq: {0}\nMem Speed: {1}\nBoard SKU: {2} {3}'
            result = result.format(self._micinfoDict[devID]['Frequency'],
                                   self._micinfoDict[devID]['GDDR Speed'],
                                   self.mic_stepping(),
                                   self._micinfoDict[devID]['Board SKU'])
        except KeyError:
            try:
                result = 'Core Freq: {0}\nMem Speed: {1}\nBoard Stepping: {2}'
                result = result.format(self._micinfoDict[devID]['Frequency'],
                                       self._micinfoDict[devID]['GDDR Speed'],
                                       self.mic_stepping())
            except KeyError:
                result = ''
        return result

    def is_same_sku(self, other):
        deprecationDict = {'B1QS-7110 P/A': 'B1QS-7110 P/A/X',
                           'C0-3120P/3120A': 'C0-3120 P/A',
                           'C0-5110P/5120D': 'C0-5110P',
                           'C0-7120P/7120X/7120': 'C0-7120 P/A/X/D'}
        mySku = deprecationDict.get(self.mic_sku(), self.mic_sku())
        mySku = mySku.replace('QS', '')
        mySku = mySku.replace('PRQ', '')

        otherSku = deprecationDict.get(other.mic_sku(), other.mic_sku())
        otherSku = otherSku.replace('QS', '')
        otherSku = otherSku.replace('PRQ', '')
        return mySku == otherSku

    def is_same_hw(self, other):
        return self.system_hw_hash() == other.system_hw_hash()

    def is_lower_mpss_version(self, other):
        if distutils.version.LooseVersion(self.mpss_version()) < other.mpss_version():
            return True
        else:
            return False

    def get_app_list(self):
        return sorted(self._commandDict.keys())

    def print_all(self):
        print self

    def system_hw_hash(self):
        micinfoDictStd = copy.deepcopy(self._micinfoDict)
        criticalKeys = ['Host Physical Memory',
                        'CPU Family',
                        'CPU Speed',
                        'Vendor ID',
                        'Device ID',
                        'SubSystem ID',
                        'Coprocessor Stepping ID',
                        'PCIe Width',
                        'PCIe Speed',
                        'PCIe Max payload size',
                        'Coprocessor Model',
                        'Coprocessor Model Ext',
                        'Coprocessor Type',
                        'Coprocessor Family',
                        'Coprocessor Family Ext',
                        'Total No of Active Cores',
                        'Frequency',
                        'GDDR Vendor',
                        'GDDR Size',
                        'GDDR Speed']

        # Standardize the output so that it is consistent between runs
        nonDecimal = re.compile(r'[^\d.]+')
        for infoDict in micinfoDictStd.values():
            if 'CPU Family' in infoDict:
                # Remove extra white space
                infoDict['CPU Family'] = ' '.join(infoDict['CPU Family'].split())

            if 'Host Physical Memory' in infoDict:
                orig = infoDict['Host Physical Memory']
                # Get rid of all non-numeric characters
                mb = float(nonDecimal.sub('', orig))
                # Get units into MB
                if 'GB' in orig or 'gb' in orig:
                    mb = mb*1000
                elif 'kB' in orig or 'kb' in orig:
                    mb = mb/1000
                # Round host memory to the nearest GB
                mb = str(int(round(mb, -3)))
                infoDict['Host Physical Memory'] = mb + ' MB'

            if 'CPU Speed' in infoDict:
                # special case for test systems to avoid sleep state issues
                if infoDict['CPU Family'] == 'GenuineIntel Family 6 Model 45 Stepping 7':
                    infoDict['CPU Speed'] = '2600 MHz'
                else:
                    orig = infoDict['CPU Speed']
                    # Get rid of all non-numeric characters
                    mhz = float(nonDecimal.sub('', orig))
                    # Get units into MHz
                    if 'GHz' in orig or 'ghz' in orig:
                        mhz = 1000*mhz
                    elif 'kHz' in orig or 'khz' in orig:
                        mhz = mhz/1000
                    # Round CPU Speed to the nearest 100 MHz
                    mhz = str(int(round(mhz, -2)))
                    infoDict['CPU Speed'] = mhz + ' MHz'

            if 'PCIe Width' in infoDict:
                orig = infoDict['PCIe Width']
                # get rid of all non-numeric characters
                chan = nonDecimal.sub('', orig)
                infoDict['PCIe Width'] = chan

            if 'PCIe Speed' in infoDict:
                orig = infoDict['PCIe Speed']
                # Get rid of all non-numeric characters
                gts = float(nonDecimal.sub('', orig))
                # Round PCIe speed to the nearest GT/s
                gts = str(int(round(gts)))
                infoDict['PCIe Speed'] = gts + ' GT/s'

            if 'Frequency' in infoDict:
                orig = infoDict['Frequency']
                # Get rid of all non-numeric characters
                khz = float(nonDecimal.sub('', orig))
                if 'MHz' in orig or 'mhz' in orig:
                    khz = 1000*khz
                elif 'GHz' in orig or 'ghz' in orig:
                    khz = 1000000*khz
                # Round Frequency to the nearest 100 MHz
                khz = str(int(round(khz, -5)))
                infoDict['Frequency'] = khz + ' kHz'

            if 'GDDR Size' in infoDict:
                orig = infoDict['GDDR Size']
                # Get rid of all non-numeric characters
                mb = float(nonDecimal.sub('', orig))
                if 'kB' in orig or 'kb' in orig:
                    mb = mb/1000
                elif 'GB' in orig or 'gb' in orig:
                    mb = 1000*mb
                # Round GDDR Size to the nearest GB
                mb = str(int(round(mb, -3)))
                infoDict['GDDR Size'] = mb + ' MB'

        everythingStr = ':'.join([':'.join([criticalKey + '=' + micinfoDictStd[devID][criticalKey]
                                            for criticalKey in criticalKeys
                                            if criticalKey in micinfoDictStd[devID]])
                                            for devID in sorted(micinfoDictStd.keys())])
        hasher = hashlib.md5()
        hasher.update(everythingStr)
        md5hash = hasher.hexdigest()
        return md5hash[:8]

    def mpss_version(self):
        return self._micinfoDict['host']['MPSS Version']

    def _init_command_list(self):
        if micp_common.is_platform_windows():
            self._commands = ['micinfo',
                              'set',
                              'systeminfo']
        else:
            self._commands = ['micinfo',
                              'dmidecode -t bios',
                              'dmidecode -t system',
                              'dmidecode -t baseboard',
                              'dmidecode -t chassis',
                              'dmidecode -t processor',
                              'dmidecode -t memory',
                              'dmidecode -t cache',
                              'dmidecode -t connector',
                              'dmidecode -t slot',
                              'micsmc --freq',
                              'micsmc --info',
                              'micsmc --mem',
                              'micsmc --temp',
                              'uname --kernel-name',
                              'uname --nodename',
                              'uname --kernel-release',
                              'uname --kernel-version',
                              'uname --machine',
                              'uname --processor',
                              'uname --hardware-platform',
                              'uname --operating-system',
                              'rpm -qa intel-mic\\* ',
                              'lspci -mmvvkD',
                              'env']

            fileName = '/etc/mpss/mic{0}.conf'
            confFiles = ['/etc/modprobe.d/mic.conf',
                         '/etc/mpss/default.conf']
            index = 0
            while os.path.exists(fileName.format(index)):
                confFiles.append(fileName.format(index))
                index = index + 1
            self._commands.extend(['egrep -v "^[ \t]*$|^#" {0}'.format(cf)
                                   for cf in confFiles])
    def _init_command_dict(self):
        self._commandDict = {}
        for cmd in self._commands:
            try:
                pid = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=-1,
                                       shell=micp_common.is_platform_windows())
                out, err = pid.communicate()
                self._commandDict[cmd] = '\n'.join((out, err))
            except OSError:
                if cmd.startswith('lspci'):
                    try:
                        cmd = '/sbin/' + cmd
                        pid = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=-1)
                        out, err = pid.communicate()
                        self._commandDict[cmd] = '\n'.join((out, err))
                    except OSError:
                        sys.stderr.write('WARNING: Could not find lspci in path, and it is not located in /sbin/lspci.\n')
                        self._commandDict[cmd] = ''
                elif cmd.startswith('/'):
                    try:
                        cmdList = shlex.split(cmd)
                        origPath = cmdList[0]
                        cmdList[0] = os.path.basename(origPath)
                        cmd = ' '.join(cmdList)
                        pid = subprocess.Popen(cmdList, stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=-1)
                        out, err = pid.communicate()
                        self._commandDict[cmd] = '\n'.join((out, err))
                    except OSError:
                        sys.stderr.write('WARNING: Could not find {0} or {1}.\n'.format(origPath, cmdList[0]))
                        self._commandDict[cmd] = ''
                elif cmd == 'set':
                    self._commandDict[cmd] = '\n'.join(sorted(['{0}={1}'.format(var, val)
                                             for (var, val) in zip(os.environ.keys(), os.environ.values())]))
                elif cmd.startswith('type '):
                    try:
                        self._commandDict[cmd] = open(cmd[5:]).read()
                    except (IOError, IndexError):
                        self._commandDict[cmd] = ''
                else:
                    sys.stderr.write('WARNING: Could not find {0}.\n'.format(cmd.split()[0]))
                    self._commandDict[cmd] = ''

    def _init_micinfo_dict(self):
        self._micinfoDict = {}
        micinfoOutput = self.__str__('micinfo')
        # Canonicalize output
        micinfoOutput = micinfoOutput.replace('MIC Processor', 'Coprocessor')
        micinfoOutput = micinfoOutput.replace('MIC Silicon', 'Coprocessor')
        micinfoOutput = micinfoOutput.replace('MIC Board', 'Coprocessor')
        micinfoOutput = micinfoOutput.replace('coprocessor', 'Coprocessor')
        micinfoOutput = micinfoOutput.replace('PCie Max payload size', 'PCIe Max payload size')
        micinfoOutput = micinfoOutput.replace('PCie Max read req size', 'PCIe Max read req size')
        micinfoOutput = micinfoOutput.replace('Subsystem ID', 'SubSystem ID')

        micInfo = micinfoOutput.split('Device No:')

        numDevices = len(micInfo) - 1
        if numDevices == 0:
            lspci = self.__str__('lspci')
            numDevices = len([ll for ll in lspci.splitlines() if 'Co-processor' in ll])
            micInfo = ['' for ii in range(numDevices+1)]

        infoDict = {}
        # on windows get host information from micinfo
        if micp_common.is_platform_windows():
            infoDict = dict([tuple([keyValue.strip() for keyValue in line.strip().split(':')])
                             for line in micInfo[0].splitlines()
                             if line.find(':') != -1 and line.find(':') == line.rfind(':')])
            if 'MPSS Version' not in infoDict:
                sys.stderr.write('WARNING:  Unknown MPSS Version, setting to 3.1\n')
                infoDict['MPSS Version'] = '3.1'
        # otherwise use standard linux routintes for host information
        else:
            infoDict['Host OS'] = platform.uname()[0]
            infoDict['OS Version'] = platform.uname()[2]
            infoDict['MPSS Version'] = self._get_mpss_version_from_rpm()
            infoDict['Host Physical Memory'] = str(self._get_memory_size_from_meminfo(-1)/(1024**2)) + ' MB'
            infoDict['CPU Family'] = self._get_cpu_family_from_cpuinfo(-1)
            infoDict['CPU Speed'] = str(self._get_cpu_mhz_from_cpuinfo(-1))

        self._micinfoDict['host'] = infoDict

        nonDecimal = re.compile(r'[^\d.]+')
        for devIdx in range(numDevices):
            infoDict = dict([tuple([keyValue.strip() for keyValue in line.strip().split(':')])
                             for line in micInfo[devIdx+1].splitlines()
                             if line.find(':') != -1 and line.find(':') == line.rfind(':')])

            if ('Vendor ID' in infoDict and
                infoDict['Vendor ID'].startswith('0x')):
                infoDict['Vendor ID'] = infoDict['Vendor ID'][2:]

            if ('Device ID' in infoDict and
                infoDict['Device ID'].startswith('0x')):
                infoDict['Device ID'] = infoDict['Device ID'][2:]

            if ('SubSystem ID' in infoDict and
                infoDict['SubSystem ID'].startswith('0x')):
                infoDict['SubSystem ID'] = infoDict['SubSystem ID'][2:]

            if ('Total No of Active Cores' not in infoDict or
                nonDecimal.sub('', infoDict['Total No of Active Cores']) == ''):
                infoDict['Total No of Active Cores'] = str(self._get_num_cores_from_cpuinfo(devIdx))

            if ('GDDR Size' not in infoDict or
                 nonDecimal.sub('', infoDict['GDDR Size']) == ''):
                infoDict['GDDR Size'] = str(self._get_memory_size_from_meminfo(devIdx)/(1024**2)) + ' MB'

            if 'GDDR Vendor' in infoDict:
                # Make sure Hynix is not HYNIX
                if infoDict['GDDR Vendor'] == 'HYNIX':
                    infoDict['GDDR Vendor'] = 'Hynix'

            if ('PCIe Width' not in infoDict or
                nonDecimal.sub('', infoDict['PCIe Width']) == ''):
                infoDict['PCIe Width'] = '16'

            if ('PCIe Speed' not in infoDict or
                 nonDecimal.sub('', infoDict['PCIe Speed']) == ''):
                infoDict['PCIe Speed'] = '5 GT/s'

            if ('PCIe Max payload size' not in infoDict or
                nonDecimal.sub('', infoDict['PCIe Max payload size']) == ''):
                if self.mpss_version().startswith('2.0'):
                    infoDict['PCIe Max payload size'] = '64 bytes'
                else:
                    infoDict['PCIe Max payload size'] = '256 bytes'

            if ('PCIe Max read req size' not in infoDict or
                nonDecimal.sub('', infoDict['PCIe Max read req size']) == ''):
                if self.mpss_version().startswith('2.0'):
                    infoDict['PCIe Max read req size'] = '128 bytes'
                else:
                    infoDict['PCIe Max read req size'] = '512 bytes'

            self._micinfoDict['mic{0}'.format(devIdx)] = infoDict


    def  _get_num_cores_from_cpuinfo(self, devIdx):
        connect = micp_connect.MPSSConnect(devIdx)
        pid = connect.Popen('cat /proc/cpuinfo')
        cpuinfo, err = pid.communicate()
        lineExpr = re.compile(r'core id\s*:\s+\d+')
        allLines = lineExpr.findall(cpuinfo)
        result = len(set([ll.split()[-1] for ll in allLines]))
        if result == 0:
            raise RuntimeError('No mic cores found')
        return result

    def  _get_cpu_mhz_from_cpuinfo(self, devIdx):
        connect = micp_connect.MPSSConnect(devIdx)
        pid = connect.Popen('cat /proc/cpuinfo')
        cpuinfo, err = pid.communicate()
        lineExpr = re.compile(r'cpu MHz\s*:\s+[\d.]+')
        allLines = lineExpr.findall(cpuinfo)
        result = str(max([float(line.split(':')[1]) for line in allLines]))
        return result

    def  _get_memory_size_from_meminfo(self, devIdx):
        connect = micp_connect.MPSSConnect(devIdx)
        pid = connect.Popen('cat /proc/meminfo')
        meminfo, err = pid.communicate()
        memDict = dict([tuple([keyValue.strip() for keyValue in line.strip().split(':')])
                        for line in meminfo.splitlines()
                        if line.find(':') != -1 and line.find(':') == line.rfind(':')])
        value, units = tuple(memDict['MemTotal'].split())
        if units.lower() == 'b':
            result = int(value)
        elif units.lower() == 'kb':
            result = int(float(value)*1024**1)
        elif units.lower() == 'mb':
            result = int(float(value)*1024**2)
        elif units.lower() == 'gb':
            result = int(float(value)*1024**3)
        else:
            raise NameError('/proc/meminfo unknown units:  {0}'.format(units))
        return result

    def _get_mpss_version_from_rpm(self):
        result = None
        cmdList = ["rpm -qa --queryformat '%{version}' mpss-modules-`uname -r`",
                   "rpm -qa --queryformat '%{version}' intel-mic"]
        for cmd in cmdList:
            pid = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            out, err = pid.communicate()
            if pid.returncode == 0 and out.strip():
                result = out.strip()
                break
        if not result:
            result = '3.1'
            sys.stderr.write("WARNING: unable to determine MPSS version from installed RPM's, setting to 3.1\n")
        return result

    def _get_cpu_family_from_cpuinfo(self, devIdx):
        connect = micp_connect.MPSSConnect(devIdx)
        pid = connect.Popen('cat /proc/cpuinfo')
        cpuinfo, err = pid.communicate()

        lineExpr = re.compile(r'vendor_id\s*:.*\n')
        allLines = lineExpr.findall(cpuinfo)
        vendorID = [line.split(':')[1].strip() for line in allLines]
        if len(set(vendorID)) > 1:
            sys.stderr.write('WARNING: multiple vendor_id values listed in /proc/cpuinfo, using only {0}\n'.format(vendorID[0]))
        vendorID = vendorID[0]

        lineExpr = re.compile(r'cpu family\s*:.*\n')
        allLines = lineExpr.findall(cpuinfo)
        cpuFamily = [line.split(':')[1].strip() for line in allLines]
        if len(set(cpuFamily)) > 1:
            sys.stderr.write('WARNING: multiple cpu family values listed in /proc/cpuinfo, using only {0}\n'.format(cpuFamily[0]))
        cpuFamily = cpuFamily[0]

        lineExpr = re.compile(r'model\s*:.*\n')
        allLines = lineExpr.findall(cpuinfo)
        model = [line.split(':')[1].strip() for line in allLines]
        if len(set(model)) > 1:
            sys.stderr.write('WARNING:  mulitple model values listed in /proc/cpuinfo, using only {0}\n'.format(model[0]))
        model = model[0]

        lineExpr = re.compile(r'stepping\s*:.*\n')
        allLines = lineExpr.findall(cpuinfo)
        stepping = [line.split(':')[1].strip() for line in allLines]
        if len(set(stepping)) > 1:
            sys.stderr.write('WARNING:  mulitple stepping values listed in /proc/cpuinfo, using only {0}\n'.format(stepping[0]))
        stepping = stepping[0]

        return '{0} Family {1} Model {2} Stepping {3}'.format(vendorID, cpuFamily, model, stepping)

