#  Copyright 2012, 2013, 2014 Intel Corporation All Rights Reserved.
#
# This software is supplied under the terms of a license
# agreement or nondisclosure agreement with Intel Corp.
# and may not be copied or disclosed except in accordance
# with the terms of that agreement.
#
#  Author:  Christopher M. Cantalupo

import os
import hashlib
import time
import re
import tempfile

import micp.kernel as micp_kernel
import micp.info as micp_info
import micp.common as micp_common

paramFileFormat = \
"""\
Generated Intel(R) LINPACK data file (lininput_xeon64)
Intel(R) LINPACK data
1
{matrixSize}
{leadDim}
{numRep}
4
"""


class linpack(micp_kernel.Kernel):
    def __init__(self):
        self.name = 'linpack'
        self._paramNames = ['omp_num_threads',
                            'matrix_size',
                            'num_rep',
                            'lead_dim']
        self._paramDefaults = {'omp_num_threads':'228',
                               'matrix_size':'8192',
                               'num_rep':'3',
                               'lead_dim':None}

        threadsPerCore = 4
        info = micp_info.Info()
        maxCount = info.num_cores()
        # limit the matrix to fit in memory with a GB to spare
        maxMemory = info.gddr_size() - 1024**3
        if maxMemory <= 0:
            raise RuntimeError('micinfo reports less than one GB of GDDR memory on card')

        self._categoryParams = {}
        # Self check always performed
        self._categoryParams['test'] = [' ']

        # Define the scaling category
        sizeConfig = [(2048, 2112),
                      (4096, 4160),
                      (6144, 6208),
                      (8192, 8256),
                      (10240, 10304),
                      (12288, 12352),
                      (14336, 14400),
                      (16384, 18496),
                      (18432, 18496),
                      (20480, 20544),
                      (22528, 22592),
                      (24576, 26688),
                      (26624, 26688),
                      (28672, 28736),
                      (30720, 30784),
                      (32768, 32832),
                      (34816, 34880),
                      (36864, 36928),
                      (38912, 38976),
                      (40960, 41024),
                      (43008, 43072),
                      (45056, 45120)]

        args = '--omp_num_threads {0} --matrix_size {1} --num_rep 3 --lead_dim {2}'
        self._categoryParams['scaling'] = [args.format(maxCount*threadsPerCore, sizes[0], sizes[1])
                                           for sizes in sizeConfig
                                           if sizes[0]*sizes[1]*8 < maxMemory]
        self._categoryParams['optimal'] = self._categoryParams['scaling'][-1:]

        self._categoryParams['scaling_quick'] = self._categoryParams['scaling'][:4]
        self._categoryParams['optimal_quick'] = self._categoryParams['scaling'][3:4]

        step = int(round(maxCount/10.0))
        if step < 4:
            step = 4
        coreConfig = range(step, maxCount, step)
        coreConfig.append(maxCount)
        self._categoryParams['scaling_core'] = [args.format(coreCount*threadsPerCore, 8192, 8256)
                                                for coreCount in coreConfig]

        self._leadDimMap = dict([(256,320), (768,832), (1024,2112), (1280,1344), (1792,1856),
                                 (2048,3136), (2304,2368), (2816,2880), (3072,4160), (3328,3392),
                                 (3840,3904), (4096,5184), (4352,4416), (4864,4928), (5120,6208),
                                 (5376,5440), (5888,5952), (6144,7232), (6400,6464), (6912,6976),
                                 (7168,8256), (7424,7488), (7936,8000), (8192,9280), (8448,8512),
                                 (8960,9024), (9216,10304), (9472,9536), (9984,10048), (10240,11328),
                                 (10496,10560), (11008,11072), (11264,12352), (11520,11584),
                                 (12032,12096), (12288,13376), (12544,12608), (13056,13120),
                                 (13312,14400), (13568,13632), (14080,14144), (14336,15424),
                                 (14592,14656), (15104,15168), (15360,16448), (15616,15680),
                                 (16128,16192), (16384,17472), (16640,16704), (17152,17216),
                                 (17408,18496), (17664,17728), (18176,18240), (18432,19520),
                                 (18688,18752), (19200,19264), (19456,20544), (19712,19776),
                                 (20224,20288), (20480,21568), (20736,20800), (21248,21312),
                                 (21504,22592), (21760,21824), (22272,22336), (22528,23616),
                                 (22784,22848), (23296,23360), (23552,24640), (23808,23872),
                                 (24320,24384), (24576,25664), (24832,24896), (25344,25408),
                                 (25600,26688), (25856,25920), (26368,26432), (26624,27712),
                                 (26880,26944), (27392,27456), (27648,28736), (27904,27968),
                                 (28416,28480), (28672,29760), (28928,28992), (29440,29504),
                                 (29696,30784), (29952,30016), (30464,30528), (30720,31808),
                                 (30976,31040), (31488,31552), (31744,32832), (32000,32064),
                                 (32512,32576), (32768,33856), (33024,33088), (33536,33600),
                                 (33792,34880), (34048,34112), (34560,34624), (34816,35904),
                                 (35072,35136), (35584,35648), (35840,36928), (36096,36160),
                                 (36608,36672), (36864,37952), (37120,37184), (37632,37696),
                                 (37888,38976), (38144,38208), (38656,38720), (38912,40000),
                                 (39168,39232), (39680,39744), (39936,41024), (40192,40256)])
        self._set_defaults_to_optimal()

    def _do_unit_test(self):
        return True

    def offload_methods(self):
        return['native', 'auto']

    def path_host_exec(self, offType):
        if offType == 'auto':
            if micp_common.is_platform_windows():
                execName = 'linpack_xeon64.exe'
            else:
                execName = 'xlinpack_xeon64'
            if 'MKLROOT' in os.environ:
                return os.path.join(os.environ['MKLROOT'], 'benchmarks', 'linpack', execName)
            else:
                raise micp_kernel.NoExecutableError("MKLROOT not in environment.  Source composer's compilervars.sh before running linpack")
        return None

    def path_dev_exec(self, offType):
        if offType == 'native':
            if 'MKLROOT' in os.environ:
                return os.path.join(os.environ['MKLROOT'], 'benchmarks', 'linpack', 'xlinpack_mic')
            else:
                raise micp_kernel.NoExecutableError("MKLROOT not in environment.  Source composer's compilervars.sh before running linpack")
        return None

    def path_aux_data(self, offType):
        result = []
        if offType == 'native':
            result.append(self.mic_library_find('libiomp5.so'))
        return result

    def param_type(self):
        return 'file'

    def parse_desc(self, raw):
        failRE = re.compile('fail', re.IGNORECASE)
        # Raise exception if self check failed
        if failRE.search(raw):
            eMessage = [line for line in raw.splitlines()
                        if failRE.search(line) or
                        line.strip().endswith('Check')]
            eMessage.insert(0, '')
            eMessage = '\n'.join(eMessage)
            raise micp_kernel.SelfCheckError(eMessage)
        # Create a dictionary from lines containing one : character
        dd = dict([tuple([ll.strip() for ll in line.split(':')])
                   for line in raw.splitlines()
                   if ':' in line and line.find(':') == line.rfind(':')])
        # Pull out important keys
        matrixSize = dd['Number of equations to solve (problem size)']
        numThreads = dd['Number of threads']
        numIt = dd['Number of trials to run']
        # Create output description
        result = '{0} x {0} MKL DP LINPACK with {1} threads and {2} iterations'.format(matrixSize, numThreads, numIt)
        return result

    def parse_perf(self, raw):
        lines = [ll.strip() for ll in raw.splitlines()]
        lineNum = lines.index('Performance Summary (GFlops)') + 3
        speed = lines[lineNum].split()[3]
        result = {}
        result['Computation.Avg'] = {'value':speed, 'units':'GFlops', 'rollup':True}
        return result

    def environment_dev(self):
        return {'LD_LIBRARY_PATH':'/tmp',
                'KMP_AFFINITY':'explicit,granularity=fine,proclist=[1-{0},0]'.format(micp_info.Info().num_cores()*4-1)}

    def environment_host(self):
        info = micp_info.Info()
        numThreads = (info.num_cores() - 1) * 4
        maxMemory = str(int((info.gddr_size() - 1024**3)/(1024**3)))
        return {'MIC_BUFFERSIZE':'256M',
                'MKL_MIC_ENABLE':'1',
                'MKL_MIC_DISABLE_HOST_FALLBACK':'1',
                'MIC_ENV_PREFIX':'MIC',
                'MIC_OMP_NUM_THREADS':str(numThreads),
                'KMP_AFFINITY':'compact,1,0',
                'MIC_KMP_AFFINITY':'explicit,granularity=fine,proclist=[1-' + str(numThreads) + ':1]',
                'MIC_USE_2MB_BUFFERS':'16K',
                'MKL_MIC_MAX_MEMORY':maxMemory + 'G'}


    def param_file(self, param):
        matrixSize = param.get_named('matrix_size')
        numRep = param.get_named('num_rep')
        leadDim = param.get_named('lead_dim')
        if leadDim == None:
            if int(matrixSize) in self._leadDimMap:
                leadDim = self._leadDimMap[int(matrixSize)]
            else:
                raise RuntimeError('linpack lead_dim not specified and matrix size not in lda table')
        paramStr = paramFileFormat.format(matrixSize=matrixSize,
                                          leadDim=leadDim,
                                          numRep=numRep)
        hasher = hashlib.md5()
        hasher.update(paramStr)
        md5hash = hasher.hexdigest()

        paramFileName = '{0}_{1}_{2}.par'.format(os.getpid(), int(1000*time.time()), md5hash[0:8])
        paramFileName = os.path.join(tempfile.gettempdir(), paramFileName)

        fid = open(paramFileName, 'w')
        fid.write(paramStr)
        fid.close()
        return paramFileName

    def independent_var(self, category):
        if category == 'scaling_core':
            return 'omp_num_threads'
        return 'matrix_size'


    def param_for_env(self):
        return ['omp_num_threads']
