#  Copyright 2012, 2013, 2014 Intel Corporation All Rights Reserved.
#
# This software is supplied under the terms of a license
# agreement or nondisclosure agreement with Intel Corp.
# and may not be copied or disclosed except in accordance
# with the terms of that agreement.
#
#  Author:  Christopher M. Cantalupo

import os
import math

import micp.kernel as micp_kernel
import micp.info as micp_info
import micp.common as micp_common

class sgemm(micp_kernel.Kernel):
    def __init__(self):
        self.name = 'sgemm'
        self._paramNames = ['f_first_matrix_size',
                            'i_num_rep',
                            'T_device',
                            'n_num_thread',
                            'm_mode',
                            'l_last_matrix_size',
                            's_step']

        info = micp_info.Info()
        devIdx = info.get_device_index()

        self._paramDefaults = {'f_first_matrix_size':'1024',
                               'i_num_rep':'3',
                               'T_device':str(devIdx),
                               'n_num_thread':'228',
                               'm_mode':'NN',
                               'l_last_matrix_size':'1024',
                               's_step':'1024'}

        self._categoryParams = {}
        self._categoryParams['test'] = [' ']
        maxCount = info.num_cores()
        maxMemory = info.gddr_size() - 1024**3
        maxMatrixSize = int(math.sqrt(maxMemory/4.0/3.0))
        if maxMatrixSize > 16384:
            maxMatrixSize = 16384
        threadsPerCore = 4

        args = '--n_num_thread {0} --f_first_matrix_size {1} --l_last_matrix_size {1} --s_step {1} --i_num_rep 3'
        maxStep = maxMatrixSize/512
        maxMatrixSize = maxStep*512
        matrixConfig = [512*ii for ii in range(1, maxStep+1)]

        self._categoryParams['scaling'] = [args.format(0, matrixSize)
                                           for matrixSize in matrixConfig]

        self._categoryParams['scaling_quick'] = [args.format(0, matrixSize)
                                                 for matrixSize in matrixConfig[:10]]

        if maxMatrixSize < 15872:
            self._categoryParams['optimal'] = [args.format(0, maxMatrixSize)]
        elif maxCount == 52:
            self._categoryParams['optimal'] = [args.format(0, 13312)]
        else:
            self._categoryParams['optimal'] = [args.format(0, 15872)]

        self._categoryParams['optimal_quick'] = [args.format(0, 5120)]

        step = int(round(maxCount/10.0))
        if step < 4:
            step = 4
        coreConfig = range(step, maxCount, step)
        coreConfig.append(maxCount)
        self._categoryParams['scaling_core'] = [args.format(coreCount*threadsPerCore, 8192)
                                                for coreCount in coreConfig]
        self._set_defaults_to_optimal()

    def _do_unit_test(self):
        return False

    def offload_methods(self):
        return['native', 'pragma', 'auto']

    def path_host_exec(self, offType):
        if offType == 'pragma':
            if micp_common.is_platform_windows():
                return self._path_exec('x86_64', 'sgemm_ofl')
            else:
                return self._path_exec('x86_64', 'sgemm_ofl.x')
        if offType == 'auto':
            if micp_common.is_platform_windows():
                return self._path_exec('x86_64', 'sgemm_cpu')
            else:
                return self._path_exec('x86_64', 'sgemm_cpu.x')
        return None

    def path_dev_exec(self, offType):
        if offType == 'native':
            return self._path_exec('k1om', 'sgemm_mic.x')
        return None

    def path_aux_data(self, offType):
        result = []
        if offType == 'native':
            result.append(self.mic_library_find('libiomp5.so'))
        return result

    def param_type(self):
        return 'flag'

    def parse_desc(self, raw):
        line = [line for line in raw.splitlines() if line.startswith('*')][0]
        matrixSize = line.split()[1]

        dd = dict([tuple([ll.strip() for ll in line.split(':')])
                   for line in raw.splitlines()
                   if ':' in line and line.find(':') == line.rfind(':')])

        numThreads = dd['num_threads']
        numIt = dd['min_niters']

        result = '{0} x {0} MKL SGEMM with {1} threads and {2} iterations'.format(matrixSize, numThreads, numIt)
        return result

    def parse_perf(self, raw):
        line = [line for line in raw.splitlines() if line.startswith('*')][0]
        speed = line.split()[3]

        dd = dict([tuple([ll.strip() for ll in line.split(':')])
                   for line in raw.splitlines()
                   if ':' in line and line.find(':') == line.rfind(':')])
        try:
            if dd['timer'] == 'native':
                tag = 'Task.Computation.Avg'
            elif dd['timer'] == 'invoke':
                tag = 'Device.Computation.Avg'
            elif dd['timer'] == 'full':
                tag = 'Host.Computation.Avg'
        except KeyError:
            tag = 'Computation.Avg'
        result = {}
        result[tag] = {'value':speed, 'units':'GFlops', 'rollup':True}
        return result

    def environment_dev(self):
        return {'LD_LIBRARY_PATH':'/tmp'}

    def environment_host(self):
        info = micp_info.Info()
        numThreads = (info.num_cores() - 1) * 4
        maxMemory = str(int((info.gddr_size() - 1024**3)/(1024**3)))
        return {'MIC_BUFFERSIZE':'256M',
                'MKL_MIC_ENABLE':'1',
                'MKL_MIC_DISABLE_HOST_FALLBACK':'1',
                'LD_LIBRARY_PATH':self.ld_library_path(),
                'MIC_LD_LIBRARY_PATH':self.mic_ld_library_path(),
                'MIC_ENV_PREFIX':'MIC',
                'MIC_OMP_NUM_THREADS':str(numThreads),
                'KMP_AFFINITY':'compact,1,0',
                'MIC_KMP_AFFINITY':'explicit,granularity=fine,proclist=[1-' + str(numThreads) + ':1]',
                'MIC_USE_2MB_BUFFERS':'16K',
                'MKL_MIC_MAX_MEMORY':maxMemory + 'G'}

    def independent_var(self, category):
        if category == 'scaling_core':
            return 'n_num_thread'
        return 'f_first_matrix_size'

    def device_param_name(self):
        return 'T_device'
